"use strict";
/**
 * Updated for TW-STYLED
 *
 * This file handles transforming template literals to class names or styled components and generates CSS content.
 * It uses CSS code from template literals and evaluated values of lazy dependencies stored in ValueCache.
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const generator_1 = __importDefault(require("@babel/generator"));
const clsx_1 = __importDefault(require("clsx"));
const logger_1 = require("@linaria/logger");
const linaria_process_css_1 = __importDefault(require("./linaria-process-css"));
const getLinariaComment = require('@linaria/babel-preset/lib/utils/getLinariaComment')
    .default;
const isSerializable = require('@linaria/babel-preset/lib/utils/isSerializable')
    .default;
const throwIfInvalid = require('@linaria/babel-preset/lib/utils/throwIfInvalid')
    .default;
const stripLines = require('@linaria/babel-preset/lib/utils/stripLines').default;
const toCSS = require('@linaria/babel-preset/lib/utils/toCSS').default;
const { units } = require('@linaria/babel-preset/lib/units');
// Match any valid CSS units followed by a separator such as ;, newline etc.
// eslint-disable-next-line @rushstack/security/no-unsafe-regexp
const unitRegex = new RegExp(`^(${units.join('|')})(;|,|\n| |\\))`);
function hasMeta(value) {
    return value && typeof value === 'object' && value.__linaria;
}
const processedPaths = new WeakSet();
function getTemplateProcessor({ types: t }, options) {
    const processCssText = linaria_process_css_1.default(options);
    return function process({ styled, path }, index, state, valueCache) {
        var _a, _b, _c;
        if (processedPaths.has(path)) {
            // Do not process an expression
            // if it is referenced in one template more than once
            return;
        }
        processedPaths.add(path);
        const { quasi } = path.node;
        const interpolations = [];
        // Check if the variable is referenced anywhere for basic DCE
        // Only works when it's assigned to a variable
        let isReferenced = true;
        const [slug, displayName, className] = getLinariaComment(path);
        const parent = path.findParent((p) => t.isObjectProperty(p) ||
            t.isJSXOpeningElement(p) ||
            t.isVariableDeclarator(p));
        if (parent) {
            const parentNode = parent.node;
            if (t.isVariableDeclarator(parentNode) && t.isIdentifier(parentNode.id)) {
                const { referencePaths } = path.scope.getBinding(parentNode.id.name) || { referencePaths: [] };
                isReferenced = referencePaths.length !== 0;
            }
        }
        // Serialize the tagged template literal to a string
        let cssText = '';
        const expressions = path.get('quasi').get('expressions');
        quasi.quasis.forEach((el, i, self) => {
            let appended = false;
            if (i !== 0 && el.value.cooked) {
                // Check if previous expression was a CSS variable that we replaced
                // If it has a unit after it, we need to move the unit into the interpolation
                // e.g. `var(--size)px` should actually be `var(--size)`
                // So we check if the current text starts with a unit, and add the unit to the previous interpolation
                // Another approach would be `calc(var(--size) * 1px), but some browsers don't support all units
                // https://bugzilla.mozilla.org/show_bug.cgi?id=956573
                const matches = el.value.cooked.match(unitRegex);
                if (matches) {
                    const last = interpolations[interpolations.length - 1];
                    const [, unit] = matches;
                    if (last && cssText.endsWith(`var(--${last.id})`)) {
                        last.unit = unit;
                        cssText += el.value.cooked.replace(unitRegex, '$2');
                        appended = true;
                    }
                }
            }
            if (!appended) {
                cssText += el.value.cooked;
            }
            const ex = expressions[i];
            if (ex) {
                const { end } = ex.node.loc;
                const result = ex.evaluate();
                const beforeLength = cssText.length;
                // The location will be end of the current string to start of next string
                const next = self[i + 1];
                const loc = {
                    // +1 because the expressions location always shows 1 column before
                    start: { line: el.loc.end.line, column: el.loc.end.column + 1 },
                    end: next
                        ? { line: next.loc.start.line, column: next.loc.start.column }
                        : { line: end.line, column: end.column + 1 }
                };
                if (result.confident) {
                    throwIfInvalid(result.value, ex);
                    if (isSerializable(result.value)) {
                        // If it's a plain object or an array, convert it to a CSS string
                        cssText += stripLines(loc, toCSS(result.value));
                    }
                    else {
                        cssText += stripLines(loc, result.value);
                    }
                    state.replacements.push({
                        original: loc,
                        length: cssText.length - beforeLength
                    });
                }
                else {
                    // Try to preval the value
                    if (options.evaluate &&
                        !(t.isFunctionExpression(ex) || t.isArrowFunctionExpression(ex))) {
                        const value = valueCache.get(ex.node);
                        throwIfInvalid(value, ex);
                        // Skip the blank string instead of throwing an error
                        if (value === '') {
                            return;
                        }
                        if (value && typeof value !== 'function') {
                            // Only insert text for non functions
                            // We don't touch functions because they'll be interpolated at runtime
                            if (hasMeta(value)) {
                                // If it's an React component wrapped in styled, get the class name
                                // Useful for interpolating components
                                cssText += `.${value.__linaria.className}`;
                            }
                            else if (isSerializable(value)) {
                                cssText += stripLines(loc, toCSS(value));
                            }
                            else {
                                // For anything else, assume it'll be stringified
                                cssText += stripLines(loc, value);
                            }
                            state.replacements.push({
                                original: loc,
                                length: cssText.length - beforeLength
                            });
                            return;
                        }
                    }
                    if (styled) {
                        const id = `${slug}-${i}`;
                        interpolations.push({
                            id,
                            node: ex.node,
                            source: ex.getSource() || generator_1.default(ex.node).code,
                            unit: ''
                        });
                        cssText += `var(--${id})`;
                    }
                    else {
                        // CSS custom properties can't be used outside components
                        throw ex.buildCodeFrameError(`The CSS cannot contain JavaScript expressions when using the 'css' tag. To evaluate the expressions at build time, pass 'evaluate: true' to the babel plugin.`);
                    }
                }
            }
        });
        let selector = `.${className}`;
        if (styled) {
            // If `styled` wraps another component and not a primitive,
            // get its class name to create a more specific selector
            // it'll ensure that styles are overridden properly
            if (options.evaluate && t.isIdentifier(styled.component.node)) {
                let value = valueCache.get(styled.component.node.name);
                while (hasMeta(value)) {
                    selector += `.${value.__linaria.className}`;
                    value = value.__linaria.extends;
                }
            }
            const props = [];
            props.push(t.objectProperty(t.identifier('name'), t.stringLiteral(displayName)));
            // If we found any interpolations, also pass them so they can be applied
            if (interpolations.length) {
                // De-duplicate interpolations based on the source and unit
                // If two interpolations have the same source code and same unit,
                // we don't need to use 2 custom properties for them, we can use a single one
                const result = {};
                interpolations.forEach((it) => {
                    const key = it.source + it.unit;
                    if (key in result) {
                        cssText = cssText.replace(`var(--${it.id})`, `var(--${result[key].id})`);
                    }
                    else {
                        result[key] = it;
                    }
                });
                props.push(t.objectProperty(t.identifier('vars'), t.objectExpression(Object.keys(result).map((key) => {
                    const { id, node, unit } = result[key];
                    const items = [node];
                    if (unit) {
                        items.push(t.stringLiteral(unit));
                    }
                    return t.objectProperty(t.stringLiteral(id), t.arrayExpression(items));
                }))));
            }
            const updatedCss = processCssText({
                selector,
                cssText,
                state,
                isReferenced
            });
            cssText = updatedCss.cssText;
            props.push(t.objectProperty(t.identifier('class'), t.stringLiteral(updatedCss.twclasses
                ? clsx_1.default(className, updatedCss.twclasses)
                : className)));
            path.replaceWith(t.callExpression(t.callExpression(t.identifier(state.file.metadata.localName || 'styled'), [styled.component.node]), [t.objectExpression(props)]));
            path.addComment('leading', '#__PURE__');
        }
        else {
            const updatedCss = processCssText({
                selector,
                cssText,
                state,
                isReferenced
            });
            cssText = updatedCss.cssText;
            path.replaceWith(t.stringLiteral(className));
        }
        if (!isReferenced && !cssText.includes(':global')) {
            return;
        }
        logger_1.debug('evaluator:template-processor:extracted-rule', `\n${selector} {${cssText}\n}`);
        const rule = {
            cssText,
            className: className,
            displayName: displayName,
            start: (_c = (_b = (_a = path.parent) === null || _a === void 0 ? void 0 : _a.loc) === null || _b === void 0 ? void 0 : _b.start) !== null && _c !== void 0 ? _c : null
        };
        state.mappings.push({
            generated: {
                line: index + 1,
                column: 0
            },
            original: rule.start,
            name: selector,
            source: `./${state.resourceFileName}`
        });
        state.cssText += `${cssText}\n`;
        state.rules[selector] = rule;
    };
}
exports.default = getTemplateProcessor;
//# sourceMappingURL=linaria-template-processor.js.map