"use strict";
/**
 * This is an entry point for styles extraction.  No logic changes from @linaria/babel-preset/src/extract,
 * except for calling our own custom template processor and setting up some global state see *** TWSTYLED ***
 *
 * On enter, It:
 *  - traverse the code using visitors (TaggedTemplateExpression, ImportDeclaration)
 *  - schedule evaluation of lazy dependencies (those who are not simple expressions //TODO does they have it's name?)
 *  - let templateProcessor to save evaluated values in babel state as `replacements`.
 * On exit, It:
 *  - store result of extraction in babel's file metadata
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const template_1 = require("@babel/template");
const generator_1 = __importDefault(require("@babel/generator"));
const logger_1 = require("@linaria/logger");
const babel_preset_1 = require("@linaria/babel-preset");
// ** TWSTYLED-BEGIN
const babel_plugin_tw_1 = require("@twstyled/babel-plugin-tw");
const visitors_1 = require("./visitors");
const linaria_template_processor_1 = __importDefault(require("./linaria-template-processor"));
const babel_visitors_write_css_1 = __importDefault(require("./babel-visitors-write-css"));
const evaluate = require('@linaria/babel-preset/lib/evaluators').default;
function isLazyValue(v) {
    return v.kind === babel_preset_1.ValueType.LAZY;
}
function isNodePath(obj) {
    return 'node' in obj && (obj === null || obj === void 0 ? void 0 : obj.node) !== undefined;
}
function findFreeName(scope, name) {
    // By default `name` is used as a name of the function …
    let nextName = name;
    let idx = 0;
    while (scope.hasBinding(nextName, false)) {
        // … but if there is an already defined variable with this name …
        // … we are trying to use a name like wrap_N
        idx += 1;
        nextName = `wrap_${idx}`;
    }
    return nextName;
}
function unwrapNode(item) {
    if (typeof item === 'string') {
        return item;
    }
    return isNodePath(item) ? item.node : item;
}
// All exported values will be wrapped with this function
const expressionWrapperTpl = template_1.statement(`
  const %%wrapName%% = (fn) => {
    try {
      return fn();
    } catch (e) {
      return e;
    }
  };
`);
const expressionTpl = template_1.expression(`%%wrapName%%(() => %%expression%%)`);
const exportsLinariaPrevalTpl = template_1.statement(`exports.__linariaPreval = %%expressions%%`);
function addLinariaPreval({ types: t }, path, lazyDeps) {
    // Constant __linariaPreval with all dependencies
    const wrapName = findFreeName(path.scope, '_wrap');
    const statements = [
        expressionWrapperTpl({ wrapName }),
        exportsLinariaPrevalTpl({
            expressions: t.arrayExpression(lazyDeps.map((expression) => expressionTpl({ expression, wrapName })))
        })
    ];
    const programNode = path.node;
    return t.program([...programNode.body, ...statements], programNode.directives, programNode.sourceType, programNode.interpreter);
}
function extract(babel, options) {
    const process = linaria_template_processor_1.default(babel, options);
    // *** TWSTYLED BEGIN
    const { visitor: visitorWriteCss } = babel_visitors_write_css_1.default(babel, options);
    const { visitor: visitorWriteTwCss } = babel_plugin_tw_1.getVisitorsWriteTwCss(babel, options);
    const { visitor: visitorsTwJsx } = babel_plugin_tw_1.getVisitorsTwJsx(babel, options);
    // *** TWSTYLED END
    return {
        visitor: {
            Program: {
                enter(path, state) {
                    // Collect all the style rules from the styles we encounter
                    state.queue = [];
                    state.rules = {};
                    state.index = -1;
                    state.dependencies = [];
                    state.replacements = [];
                    // *** TWSTYLED BEGIN
                    visitorWriteCss.Program.enter(path, state);
                    visitorWriteTwCss.Program.enter(path, state);
                    // *** TWSTYLED END
                    logger_1.debug('extraction:start', state.file.opts.filename);
                    // Invalidate cache for module evaluation to get fresh modules
                    babel_preset_1.Module.invalidate();
                    // We need our transforms to run before anything else
                    // So we traverse here instead of a in a visitor
                    path.traverse({
                        ImportDeclaration: (p) => visitors_1.DetectStyledImportName(babel, p, state),
                        TaggedTemplateExpression: (p) => {
                            visitors_1.GenerateClassNames(babel, p, state, options);
                            visitors_1.CollectDependencies(babel, p, state, options);
                        }
                    });
                    const lazyDeps = state.queue.reduce((acc, { expressionValues: values }) => {
                        acc.push(...values.filter(isLazyValue));
                        return acc;
                    }, []);
                    const expressionsToEvaluate = lazyDeps.map((v) => unwrapNode(v.ex));
                    const originalLazyExpressions = lazyDeps.map((v) => unwrapNode(v.originalEx));
                    logger_1.debug('lazy-deps:count', lazyDeps.length);
                    let lazyValues = [];
                    if (expressionsToEvaluate.length > 0) {
                        logger_1.debug('lazy-deps:original-expressions-list', originalLazyExpressions.map((node) => typeof node !== 'string' ? generator_1.default(node).code : node));
                        logger_1.debug('lazy-deps:expressions-to-eval-list', expressionsToEvaluate.map((node) => typeof node !== 'string' ? generator_1.default(node).code : node));
                        const program = addLinariaPreval(babel, path, expressionsToEvaluate);
                        const { code } = generator_1.default(program);
                        logger_1.debug('lazy-deps:evaluate', '');
                        try {
                            const evaluation = evaluate(code, state.file.opts.filename, options);
                            logger_1.debug('lazy-deps:sub-files', evaluation.dependencies);
                            state.dependencies.push(...evaluation.dependencies);
                            lazyValues = evaluation.value.__linariaPreval || [];
                            logger_1.debug('lazy-deps:values', evaluation.value.__linariaPreval);
                        }
                        catch (e) {
                            throw new Error('An unexpected runtime error occurred during dependencies evaluation: \n' +
                                e.stack +
                                '\n\nIt may happen when your code or third party module is invalid or uses identifiers not available in Node environment, eg. window. \n' +
                                'Note that line numbers in above stack trace will most likely not match, because Linaria needed to transform your code a bit.\n');
                        }
                    }
                    const valueCache = new Map();
                    originalLazyExpressions.forEach((key, idx) => valueCache.set(key, lazyValues[idx]));
                    state.queue.forEach((item, index) => process(item, index, state, valueCache));
                },
                exit(nodePath, state) {
                    if (Object.keys(state.rules).length) {
                        // Store the result as the file metadata under linaria key
                        state.file.metadata.linaria = {
                            rules: state.rules,
                            replacements: state.replacements,
                            dependencies: state.dependencies
                        };
                    }
                    // Invalidate cache for module evaluation when we're done
                    babel_preset_1.Module.invalidate();
                    logger_1.debug('extraction:end', state.file.opts.filename);
                    // *** TWSTYLED BEGIN
                    visitorWriteCss.Program.exit(nodePath, state);
                    visitorWriteTwCss.Program.exit(nodePath, state);
                    // *** TWSTYLED END
                }
            },
            // *** TWSTYLED BEGIN
            ImportDefaultSpecifier(path, state) {
                visitorsTwJsx.ImportDefaultSpecifier(path, state);
            },
            JSXAttribute(path, state) {
                visitorsTwJsx.JSXAttribute(path, state);
            }
            // *** TWSTYLED END
        }
    };
}
exports.default = extract;
//# sourceMappingURL=plugin.js.map